@@ -27,6 +27,8 @@ import (
 	"crypto/ecdh"
 	"crypto/ecdsa"
 	"crypto/elliptic"
+	"crypto/rsa"
+	"crypto/sha1" //nolint:gosec // required to calculate hash
 	"crypto/x509"
 	"encoding/hex"
 	"errors"
@@ -68,9 +70,10 @@ func (k *keyAttributes) retryAttributes() *keyAttributes {
 		return nil
 	}
 	return &keyAttributes{
-		label: k.label,
-		hash:  k.hash,
-		retry: false,
+		label:            k.label,
+		hash:             k.hash,
+		useSecureEnclave: k.useSecureEnclave,
+		retry:            false,
 	}
 }
 
@@ -83,8 +86,10 @@ type keySearchAttributes struct {
 }
 
 type certAttributes struct {
-	label        string
-	serialNumber *big.Int
+	label                     string
+	serialNumber              *big.Int
+	keychain                  string
+	useDataProtectionKeychain bool
 }
 
 type algorithmAttributes struct {
@@ -136,11 +141,29 @@ var signatureAlgorithmMapping = map[apiv1.SignatureAlgorithm]algorithmAttributes
 //     DER representation of an RSA public key using the PKCS #1 format or the
 //     SHA-1 of the uncompressed ECDSA point according to SEC 1, Version 2.0,
 //     Section 2.3.4.
-type MacKMS struct{}
+type MacKMS struct {
+	useDataProtectionKeychain bool
+}
 
-// New returns a new SoftKMS.
-func New(context.Context, apiv1.Options) (*MacKMS, error) {
-	return &MacKMS{}, nil
+// New returns a new [MacKMS] instance. In code-signed applications, it is
+// possible to initialize [MacKMS] with a URI that specifies whether to use the
+// Data Protection Keychain for loading and storing certificates by default.
+//
+// The [New] function accepts the following URI format:
+//   - mackms:keychain=dataProtection
+func New(_ context.Context, opts apiv1.Options) (*MacKMS, error) {
+	var useDataProtectionKeychain bool
+	if opts.URI != "" {
+		u, err := uri.ParseWithScheme(Scheme, opts.URI)
+		if err != nil {
+			return nil, err
+		}
+		useDataProtectionKeychain = isDataProtectionKeychain(u.Get("keychain"), false)
+	}
+
+	return &MacKMS{
+		useDataProtectionKeychain: useDataProtectionKeychain,
+	}, nil
 }
 
 func init() {
@@ -357,18 +380,26 @@ func (k *MacKMS) CreateSigner(req *apiv1.CreateSignerRequest) (crypto.Signer, er
 //   - mackms:label=test@example.com
 //   - mackms:serial=2c273934eda8454d2595a94497e2395a
 //   - mackms:label=test@example.com;serial=2c273934eda8454d2595a94497e2395a
+//
+// On code-signed applications, it is possible to use the Data Protection
+// Keychain by default if [UseDataProtectionKeychain] is set to true. You can
+// also select the keychain using the "keychain" attribute:
+//   - "mackms:label=my-label;keychain=dataProtection"
+//   - "mackms:label=my-label;keychain=login"
+//
+// Currently, only the keychains dataProtection and login are supported.
 func (k *MacKMS) LoadCertificate(req *apiv1.LoadCertificateRequest) (*x509.Certificate, error) {
 	if req.Name == "" {
 		return nil, fmt.Errorf("loadCertificateRequest 'name' cannot be empty")
 	}
 
 	// Require label or serial
-	u, err := parseCertURI(req.Name, true)
+	u, err := parseCertURI(req.Name, k.useDataProtectionKeychain, true)
 	if err != nil {
 		return nil, fmt.Errorf("mackms LoadCertificate failed: %w", err)
 	}
 
-	cert, err := loadCertificate(u.label, u.serialNumber, nil)
+	cert, err := loadCertificate(u, nil)
 	if err != nil {
 		return nil, fmt.Errorf("mackms LoadCertificate failed: %w", apiv1Error(err))
 	}
@@ -386,6 +417,14 @@ func (k *MacKMS) LoadCertificate(req *apiv1.LoadCertificateRequest) (*x509.Certi
 //   - "mackms:" will use the common name
 //   - "mackms:label=my-label" will use "my-label"
 //   - "mackms:my-label" will use the "my-label"
+//
+// On code-signed applications, it is possible to use the Data Protection
+// Keychain by default if [UseDataProtectionKeychain] is set to true. You can
+// also select the keychain using the "keychain" attribute:
+//   - "mackms:label=my-label;keychain=dataProtection"
+//   - "mackms:label=my-label;keychain=login"
+//
+// Currently, only the keychains dataProtection and login are supported.
 func (k *MacKMS) StoreCertificate(req *apiv1.StoreCertificateRequest) error {
 	// There's not really need to require the name as macOS will use the common
 	// name as default.
@@ -395,13 +434,24 @@ func (k *MacKMS) StoreCertificate(req *apiv1.StoreCertificateRequest) error {
 
 	// Do not require any parameter. Using mackms: is allowed as macOS will set
 	// the commonName as label.
-	u, err := parseCertURI(req.Name, false)
+	u, err := parseCertURI(req.Name, k.useDataProtectionKeychain, false)
 	if err != nil {
 		return fmt.Errorf("mackms StoreCertificate failed: %w", err)
 	}
 
+	// Write the certificate in the Data Protection Keychain if the key is in
+	// the secure enclave.
+	if u.keychain == "" && !u.useDataProtectionKeychain {
+		if attrs, err := getKeyAttributes(req.Certificate.PublicKey); err == nil {
+			defer attrs.Release()
+			if security.GetSecAttrTokenID(attrs) == "com.apple.setoken" {
+				u.useDataProtectionKeychain = true
+			}
+		}
+	}
+
 	// Store the certificate and update the label if required
-	if err := storeCertificate(u.label, req.Certificate); err != nil {
+	if err := storeCertificate(u, req.Certificate); err != nil {
 		return fmt.Errorf("mackms StoreCertificate failed: %w", apiv1Error(err))
 	}
 
@@ -422,12 +472,12 @@ func (k *MacKMS) LoadCertificateChain(req *apiv1.LoadCertificateChainRequest) ([
 	}
 
 	// Require label or serial
-	u, err := parseCertURI(req.Name, true)
+	u, err := parseCertURI(req.Name, k.useDataProtectionKeychain, true)
 	if err != nil {
 		return nil, fmt.Errorf("mackms LoadCertificateChain failed: %w", err)
 	}
 
-	cert, err := loadCertificate(u.label, u.serialNumber, nil)
+	cert, err := loadCertificate(u, nil)
 	if err != nil {
 		return nil, fmt.Errorf("mackms LoadCertificateChain failed: %w", apiv1Error(err))
 	}
@@ -438,12 +488,13 @@ func (k *MacKMS) LoadCertificateChain(req *apiv1.LoadCertificateChainRequest) ([
 	}
 
 	// Look for the rest of intermediates skipping the root.
+	chainURI := &certAttributes{useDataProtectionKeychain: u.useDataProtectionKeychain}
 	for {
 		// The Keychain stores the subject as an attribute, but it saves some of
 		// the values in uppercase. We cannot use the cert.RawIssuer to restrict
 		// more the search with KSecAttrSubjectKeyID and kSecAttrSubject. To do
 		// it we will need to "normalize" the subject it in the same way.
-		parent, err := loadCertificate("", nil, cert.AuthorityKeyId)
+		parent, err := loadCertificate(chainURI, cert.AuthorityKeyId)
 		if err != nil || isSelfSigned(parent) || cert.CheckSignatureFrom(parent) != nil {
 			break
 		}
@@ -474,19 +525,31 @@ func (k *MacKMS) StoreCertificateChain(req *apiv1.StoreCertificateChainRequest)
 
 	// Do not require any parameter. Using mackms: is allowed as macOS will set
 	// the commonName as label.
-	u, err := parseCertURI(req.Name, false)
+	u, err := parseCertURI(req.Name, k.useDataProtectionKeychain, false)
 	if err != nil {
 		return fmt.Errorf("mackms StoreCertificateChain failed: %w", err)
 	}
 
+	// Write the certificate in the Data Protection Keychain if the key is in
+	// the secure enclave.
+	if u.keychain == "" && !u.useDataProtectionKeychain {
+		if attrs, err := getKeyAttributes(req.CertificateChain[0].PublicKey); err == nil {
+			defer attrs.Release()
+			if security.GetSecAttrTokenID(attrs) == "com.apple.setoken" {
+				u.useDataProtectionKeychain = true
+			}
+		}
+	}
+
 	// Store the certificate and update the label if required
-	if err := storeCertificate(u.label, req.CertificateChain[0]); err != nil {
+	if err := storeCertificate(u, req.CertificateChain[0]); err != nil {
 		return fmt.Errorf("mackms StoreCertificateChain failed: %w", apiv1Error(err))
 	}
 
 	// Store the rest of the chain but do not fail if already exists
+	chainURI := &certAttributes{useDataProtectionKeychain: u.useDataProtectionKeychain}
 	for _, cert := range req.CertificateChain[1:] {
-		if err := storeCertificate("", cert); err != nil && !errors.Is(err, security.ErrAlreadyExists) {
+		if err := storeCertificate(chainURI, cert); err != nil && !errors.Is(err, security.ErrAlreadyExists) {
 			return fmt.Errorf("mackms StoreCertificateChain failed: %w", err)
 		}
 	}
@@ -530,6 +593,11 @@ func (*MacKMS) DeleteKey(req *apiv1.DeleteKeyRequest) error {
 			defer cfTag.Release() //nolint:gocritic // only two iterations
 			dict[security.KSecAttrApplicationTag] = cfTag
 		}
+		if u.useSecureEnclave {
+			dict[security.KSecUseDataProtectionKeychain] = cf.True
+		} else {
+			dict[security.KSecUseDataProtectionKeychain] = cf.False
+		}
 		// Extract logic to deleteItem to avoid defer on loops
 		if err := deleteItem(dict, u.hash); err != nil {
 			return fmt.Errorf("mackms DeleteKey failed: %w", apiv1Error(err))
@@ -546,18 +614,19 @@ func (*MacKMS) DeleteKey(req *apiv1.DeleteKeyRequest) error {
 //
 // Notice: This API is EXPERIMENTAL and may be changed or removed in a later
 // release.
-func (*MacKMS) DeleteCertificate(req *apiv1.DeleteCertificateRequest) error {
+func (k *MacKMS) DeleteCertificate(req *apiv1.DeleteCertificateRequest) error {
 	if req.Name == "" {
 		return fmt.Errorf("deleteCertificateRequest 'name' cannot be empty")
 	}
 
-	u, err := parseCertURI(req.Name, true)
+	u, err := parseCertURI(req.Name, k.useDataProtectionKeychain, true)
 	if err != nil {
 		return fmt.Errorf("mackms DeleteCertificate failed: %w", err)
 	}
 
 	query := cf.Dictionary{
-		security.KSecClass: security.KSecClassCertificate,
+		security.KSecClass:      security.KSecClassCertificate,
+		security.KSecMatchLimit: security.KSecMatchLimitOne,
 	}
 	if u.label != "" {
 		cfLabel, err := cf.NewString(u.label)
@@ -575,6 +644,11 @@ func (*MacKMS) DeleteCertificate(req *apiv1.DeleteCertificateRequest) error {
 		defer cfSerial.Release()
 		query[security.KSecAttrSerialNumber] = cfSerial
 	}
+	if u.useDataProtectionKeychain {
+		query[security.KSecUseDataProtectionKeychain] = cf.True
+	} else {
+		query[security.KSecUseDataProtectionKeychain] = cf.False
+	}
 
 	if err := deleteItem(query, nil); err != nil {
 		return fmt.Errorf("mackms DeleteCertificate failed: %w", apiv1Error(err))
@@ -672,12 +746,12 @@ var _ apiv1.SearchableKeyManager = (*MacKMS)(nil)
 
 func deleteItem(dict cf.Dictionary, hash []byte) error {
 	if len(hash) > 0 {
-		d, err := cf.NewData(hash)
+		cfHash, err := cf.NewData(hash)
 		if err != nil {
 			return err
 		}
-		defer d.Release()
-		dict[security.KSecAttrApplicationLabel] = d
+		defer cfHash.Release()
+		dict[security.KSecAttrApplicationLabel] = cfHash
 	}
 
 	query, err := cf.NewDictionary(dict)
@@ -697,19 +771,20 @@ func deleteItem(dict cf.Dictionary, hash []byte) error {
 }
 
 func getPrivateKey(u *keyAttributes) (*security.SecKeyRef, error) {
-	cfLabel, err := cf.NewString(u.label)
-	if err != nil {
-		return nil, err
-	}
-	defer cfLabel.Release()
-
 	dict := cf.Dictionary{
 		security.KSecClass:        security.KSecClassKey,
-		security.KSecAttrLabel:    cfLabel,
 		security.KSecAttrKeyClass: security.KSecAttrKeyClassPrivate,
 		security.KSecReturnRef:    cf.True,
 		security.KSecMatchLimit:   security.KSecMatchLimitOne,
 	}
+	if u.label != "" {
+		cfLabel, err := cf.NewString(u.label)
+		if err != nil {
+			return nil, err
+		}
+		defer cfLabel.Release()
+		dict[security.KSecAttrLabel] = cfLabel
+	}
 	if u.tag != "" {
 		cfTag, err := cf.NewData([]byte(u.tag))
 		if err != nil {
@@ -726,6 +801,11 @@ func getPrivateKey(u *keyAttributes) (*security.SecKeyRef, error) {
 		defer d.Release()
 		dict[security.KSecAttrApplicationLabel] = d
 	}
+	if u.useSecureEnclave {
+		dict[security.KSecUseDataProtectionKeychain] = cf.True
+	} else {
+		dict[security.KSecUseDataProtectionKeychain] = cf.False
+	}
 
 	// Get the query from the keychain
 	query, err := cf.NewDictionary(dict)
@@ -811,6 +891,40 @@ func getPrivateKeys(u *keySearchAttributes) ([]*security.SecKeychainItemRef, err
 	return keys, nil
 }
 
+func getKeyAttributes(pub crypto.PublicKey) (*cf.DictionaryRef, error) {
+	hash, err := createHash(pub)
+	if err != nil {
+		return nil, err
+	}
+
+	cfHash, err := cf.NewData(hash)
+	if err != nil {
+		return nil, err
+	}
+	defer cfHash.Release()
+
+	query, err := cf.NewDictionary(cf.Dictionary{
+		security.KSecClass:                security.KSecClassKey,
+		security.KSecAttrKeyClass:         security.KSecAttrKeyClassPrivate,
+		security.KSecReturnRef:            cf.True,
+		security.KSecMatchLimit:           security.KSecMatchLimitOne,
+		security.KSecAttrApplicationLabel: cfHash,
+	})
+	if err != nil {
+		return nil, err
+	}
+	defer query.Release()
+
+	var ref cf.TypeRef
+	err = security.SecItemCopyMatching(query, &ref)
+	if err != nil {
+		return nil, err
+	}
+	defer ref.Release()
+
+	return security.SecKeyCopyAttributes(security.NewSecKeyRef(ref)), nil
+}
+
 func extractPublicKey(secKeyRef *security.SecKeyRef) (crypto.PublicKey, []byte, error) {
 	// Get the hash of the public key. We can also calculate this from the
 	// external representation below, but in case Apple decides to switch from
@@ -896,20 +1010,22 @@ func isSelfSigned(cert *x509.Certificate) bool {
 	return false
 }
 
-func loadCertificate(label string, serialNumber *big.Int, subjectKeyID []byte) (*x509.Certificate, error) {
+func loadCertificate(u *certAttributes, subjectKeyID []byte) (*x509.Certificate, error) {
 	query := cf.Dictionary{
-		security.KSecClass: security.KSecClassCertificate,
+		security.KSecClass:      security.KSecClassCertificate,
+		security.KSecMatchLimit: security.KSecMatchLimitOne,
+		security.KSecReturnRef:  cf.True,
 	}
-	if label != "" {
-		cfLabel, err := cf.NewString(label)
+	if u.label != "" {
+		cfLabel, err := cf.NewString(u.label)
 		if err != nil {
 			return nil, err
 		}
 		defer cfLabel.Release()
 		query[security.KSecAttrLabel] = cfLabel
 	}
-	if serialNumber != nil {
-		cfSerial, err := cf.NewData(encodeSerialNumber(serialNumber))
+	if u.serialNumber != nil {
+		cfSerial, err := cf.NewData(encodeSerialNumber(u.serialNumber))
 		if err != nil {
 			return nil, err
 		}
@@ -924,6 +1040,13 @@ func loadCertificate(label string, serialNumber *big.Int, subjectKeyID []byte) (
 		defer cfSubjectKeyID.Release()
 		query[security.KSecAttrSubjectKeyID] = cfSubjectKeyID
 	}
+	// Apple recommends to set this parameters always to true for all keychain
+	// operations. This requires a code-signed application.
+	if u.useDataProtectionKeychain {
+		query[security.KSecUseDataProtectionKeychain] = cf.True
+	} else {
+		query[security.KSecUseDataProtectionKeychain] = cf.False
+	}
 
 	cfQuery, err := cf.NewDictionary(query)
 	if err != nil {
@@ -951,7 +1074,7 @@ func loadCertificate(label string, serialNumber *big.Int, subjectKeyID []byte) (
 	return cert, nil
 }
 
-func storeCertificate(label string, cert *x509.Certificate) error {
+func storeCertificate(u *certAttributes, cert *x509.Certificate) error {
 	cfData, err := cf.NewData(cert.Raw)
 	if err != nil {
 		return err
@@ -966,10 +1089,29 @@ func storeCertificate(label string, cert *x509.Certificate) error {
 
 	// Adding the label here doesn't have any effect. Apple Keychain always uses
 	// the commonName.
-	attributes, err := cf.NewDictionary(cf.Dictionary{
+	dict := cf.Dictionary{
 		security.KSecClass:    security.KSecClassCertificate,
 		security.KSecValueRef: certRef,
-	})
+	}
+	// Apple recommends to set KSecUseDataProtectionKeychain parameters always
+	// to true for all keychain operations. This option requires to use a
+	// code-signed application.
+	if u.useDataProtectionKeychain {
+		dict[security.KSecUseDataProtectionKeychain] = cf.True
+		// The data in the keychain item cannot be accessed after a restart
+		// until the device has been unlocked once by the user.
+		access, err := security.SecAccessControlCreateWithFlags(
+			security.KSecAttrAccessibleAfterFirstUnlock,
+			0,
+		)
+		if err != nil {
+			return err
+		}
+		defer access.Release()
+		dict[security.KSecAttrAccessControl] = access
+	}
+
+	attributes, err := cf.NewDictionary(dict)
 	if err != nil {
 		return err
 	}
@@ -981,8 +1123,8 @@ func storeCertificate(label string, cert *x509.Certificate) error {
 	}
 
 	// Update the label if necessary
-	if label != "" && label != cert.Subject.CommonName {
-		cfLabel, err := cf.NewString(label)
+	if u.label != "" && u.label != cert.Subject.CommonName {
+		cfLabel, err := cf.NewString(u.label)
 		if err != nil {
 			return err
 		}
@@ -1061,7 +1203,7 @@ func parseURI(rawuri string) (*keyAttributes, error) {
 	}, nil
 }
 
-func parseCertURI(rawuri string, requireValue bool) (*certAttributes, error) {
+func parseCertURI(rawuri string, useDataProtectionKeychain, requireValue bool) (*certAttributes, error) {
 	// When rawuri is just the label
 	if !strings.HasPrefix(strings.ToLower(rawuri), Scheme) {
 		return &certAttributes{
@@ -1089,6 +1231,7 @@ func parseCertURI(rawuri string, requireValue bool) (*certAttributes, error) {
 	// With regular values, uris look like this:
 	// mackms:label=my-cert;serial=01020A0B...
 	label := u.Get("label")
+	keychain := u.Get("keychain")
 	serial := u.GetEncoded("serial")
 	if requireValue && label == "" && len(serial) == 0 {
 		return nil, fmt.Errorf("error parsing %q: label or serial are required", rawuri)
@@ -1100,8 +1243,10 @@ func parseCertURI(rawuri string, requireValue bool) (*certAttributes, error) {
 	}
 
 	return &certAttributes{
-		label:        label,
-		serialNumber: serialNumber,
+		label:                     label,
+		serialNumber:              serialNumber,
+		useDataProtectionKeychain: isDataProtectionKeychain(keychain, useDataProtectionKeychain),
+		keychain:                  keychain,
 	}, nil
 }
 
@@ -1148,6 +1293,19 @@ func parseSearchURI(rawuri string) (*keySearchAttributes, error) {
 	}, nil
 }
 
+func isDataProtectionKeychain(s string, defaultValue bool) bool {
+	switch strings.ToLower(s) {
+	case "dataprotection":
+		return true
+	case "login":
+		return false
+	case "":
+		return defaultValue
+	default:
+		return false
+	}
+}
+
 // encodeSerialNumber encodes the serial number of a certificate in ASN.1.
 // Negative serial numbers are not allowed.
 func encodeSerialNumber(s *big.Int) []byte {
@@ -1238,6 +1396,32 @@ func ecdhToECDSAPublicKey(key *ecdh.PublicKey) (*ecdsa.PublicKey, error) {
 	}
 }
 
+// createHash creates the SHA-1 of the DER representation of an RSA public key
+// using the PKCS #1 format or the SHA-1 of the uncompressed ECDSA point
+// according to SEC 1, Version 2.0, Section 2.3.4. It corresponds to the
+// kSecAttrApplicationLabel attribute
+func createHash(key crypto.PublicKey) ([]byte, error) {
+	switch k := key.(type) {
+	case *ecdsa.PublicKey:
+		pub, err := k.ECDH()
+		if err != nil {
+			return nil, err
+		}
+		return sha1Sum(pub.Bytes()), nil
+	case *rsa.PublicKey:
+		return sha1Sum(x509.MarshalPKCS1PublicKey(k)), nil
+	default:
+		return nil, fmt.Errorf("usupported public key type %T", key)
+	}
+}
+
+//nolint:gosec // required to calculate hash
+func sha1Sum(b []byte) []byte {
+	h := sha1.New()
+	h.Write(b)
+	return h.Sum(nil)
+}
+
 func apiv1Error(err error) error {
 	switch {
 	case errors.Is(err, security.ErrNotFound):
