@@ -96,8 +96,8 @@ func createIssuerCertificate(t *testing.T, commonName string) (*x509.Certificate
 		NotAfter:              now.Add(time.Hour),
 		KeyUsage:              x509.KeyUsageCertSign | x509.KeyUsageCRLSign,
 		BasicConstraintsValid: true,
-		MaxPathLen:            0,
-		MaxPathLenZero:        true,
+		MaxPathLen:            -1,
+		MaxPathLenZero:        false,
 		Issuer:                pkix.Name{CommonName: "issuer"},
 		Subject:               pkix.Name{CommonName: "issuer"},
 		SerialNumber:          sn,
